/***************************************************************************
* blitz/alg/elmhes.cc     
* function implementation of vectorized hessenberg transformation
*
* $Id: elmhes.cc,v 1.1.1.1 2003/07/09 23:55:52 mdsuresh Exp $
*
* Copyright (C) 2003 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
* 
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* Suggestions:          blitz-dev@oonumerics.org
* Bugs:                 blitz-bugs@oonumerics.org
*
* For more information, please see the Blitz++ Home Page:
*    http://oonumerics.org/blitz/
*
***************************************************************************
* $Log: elmhes.cc,v $
* Revision 1.1.1.1  2003/07/09 23:55:52  mdsuresh
*
*
* Revision 1.0  2002/10/26 20:22:49  mdsuresh
* Updated copyright date in headers.
**/


#ifndef _VT_FUNCTION_ELMHES_CC
#define _VT_FUNCTION_ELMHES_CC

#ifndef _VT_PACK_H
#error the implementation of Hessenberg reduction requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif   


namespace _VT_Pack {
	namespace Functions {
		namespace Matrix {
			namespace EigenSystems {
				template<class T>   
					void Reduce2Hessenberg(T& a)
				{
					int m,j,i;
					_bz_use(j);
					int n = _VT_rows(a);
					_VT_PRECHECK( _VT_rows(a) == _VT_cols(a),
						"This operation is only valid on square matrices.\n"
						"Invalid shape:" << _VT_shape(a) );
					_VT_numtype(T) x;
					_VT_Array1D( _VT_numtype( T ) ) y(n), tmp(n);
					_VT_DeclareRangeAll( _VT_RangeAll );

					for (m=1;m<n-1;m++) {
						tmp = _VT_abs(a(_VT_RangeAll, m-1));
                        						

						//cout << tmp << "\n";
						i = _VT_maxIndex_1( tmp(_VT_Range(m,n-1))  ,0) + m;

						//cout << "i=" << i  << "m=" << m << "\n";
						x=a(i,m-1);
						//_VT_Range all;

						if (i != m) {
							_VT_Range _j(m-1,n-1);

						/*	for(int k=m-1; k < n; k++) {
							    _VT_numtype(T) temp =a(m,k);
								a(m,k) = a(i,k);
								a(i,k) = temp;
							}
							*/

							_VT_swap(a(i,_j), a(m,_j));
						/*	for(int k=0; k < n; k++) {
							    _VT_numtype(T) temp =a(k,i);
								a(k,i) = a(k,m);
								a(k,m) = temp;
							}
							*/
							_VT_swap(a(_VT_RangeAll,i), a( _VT_RangeAll,m)); 
						}
						if (x) {
							_VT_Range q(m+1,n-1), p(m,n-1);
							y(q) = a(q,m-1)/x;
							a(q,m-1) = y(q);
							a(q,p) = a(q,p) - y(q)(_VT_tensor::i) * a(m,p)(_VT_tensor::j);
							a(_VT_RangeAll,m) = a(_VT_RangeAll,m) +
								_VT_sum(a(_VT_RangeAll,q)(_VT_tensor::i, _VT_tensor::j) 
								* y(q)(_VT_tensor::j), _VT_tensor::j);

						}
					}     


				}
			}
		}
	}    
}

#ifdef _VT_MATLAB_INTERFACE

	namespace _VT_MATLAB_NAMESPACE {
	using namespace _VT_Pack;

		/* hess: MATLAB look alike */   
		template<class T>
			T hess( T & _arr)  {
				T arr = _VT_copy(_arr) ;
				_VT_Pack::Functions::Matrix::EigenSystems::Reduce2Hessenberg(arr);
				return arr;           

			}

	}

#endif // _VT_MATLAB_INTERFACE


#endif   // _BZ_NAMESPACE_ELMHES_CC
