/***************************************************************************
* blitz/alg/fwt.cc     
* function implementation of vectorized Fast Wavelet Transformation
*
* $Id: fwt.cc,v 1.1.1.1 2003/07/09 23:55:52 mdsuresh Exp $
*
* Copyright (C) 2003 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
*
* FWT statistical package
* Copyright (C) 1990-2001 Stanford 
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* Suggestions:          blitz-dev@oonumerics.org
* Bugs:                 blitz-bugs@oonumerics.org
*
* For more information, please see the Blitz++ Home Page:
*    http://oonumerics.org/blitz/
*
***************************************************************************
* $Log: fwt.cc,v $
* Revision 1.1.1.1  2003/07/09 23:55:52  mdsuresh
*
*
* Revision 1.0  2002/10/26 09:22:49  mdsuresh
* Updated copyright date in headers.
**/


#ifndef _VT_FUNCTION_FWT_CC
#define _VT_FUNCTION_FWT_CC

#ifndef _VT_PACK_H
#error the implementation of Fast Wavelet Transform requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif   



namespace _VT_Pack {
	namespace Functions {
		namespace Spectral {
			namespace Wavelet {

				template<class T>
					T  MirrorFilter(T input) 
				{
					int s = _VT_rows(input);
					_VT_PRECHECK( _VT_rank(input) == 1,
						"Construction of wavelet filter with rank higher than 1 is "
						"currently not supported.");
					_VT_Range a(0,s-1,2);
					_VT_Range b(1,s-1,2);
					T result(_VT_shape(input));
					//  result = 0;
					result(a) = input(a);
					result(b) = -input(b);
					return result;
				}

				template<class T1, class T2>
					void DownLo(T1  x,   T2 lpf, int n, T1  result) 
				{

					int n2, mlo, mhi,m,  h, j; 
					n2 =  n/2;
					m = _VT_rows(lpf);
					mlo = m/2; 
					mhi = n2-mlo;
					if(2*mhi + (m-1) >= n) --mhi;
					_VT_DeclareRangeAll( _VT_RangeAll);

					if(mhi < 0) mhi = -1;

					else {

						_VT_Range i_read(2*0, 2*mhi,2);
						_VT_Range i_write(0, mhi);
						result(i_write) = 0.;

						for( h=0; h < m; h++)
							result(i_write) += lpf(h)*x(i_read+h);
					}


					/* fix up edge values */
					for(int i= mhi+1 ; i<n2; i++){
						_VT_Range _i(i);
						result(_i) = 0.;
						j = 2*i;
						for( h=0; h < m; h++){
							if(j >= n) j -= n;
							_VT_Range _j(j);
							result(_i) += lpf(h)*x(_j);
							j++;
						}

					}



				}



				template<class T1,class T2>
					void DownHi(T1  x, T2 hpf, int n, T1 result)

				{
					int m,n2, mlo,  h, j; 
					/* highpass version */
					n2 = n/2;
					m = _VT_rows(hpf);
					mlo = m/2-1; 
					_VT_DeclareRangeAll( _VT_RangeAll);

					if(2*mlo+1  - (m-1) < 0) mlo++;



					_VT_Range i_read( 2*(mlo) + 1, 2*(n2-1) +1,2);
					_VT_Range i_write(mlo , n2 -1); 

					//result( i_write, all) = sum(hpf(tensor::k)*x(tensor::i , tensor::j), tensor::k);

					result( i_write) = 0.;
					for( h=0; h < m; h++)
						result(i_write) += hpf(h)*x(i_read-h);



					if(mlo > n2) mlo = n2;
					/* fix up edge values */
					for(int i= 0; i<mlo; i++) {
						_VT_Range _i(i);

						result(_i) = 0.;

						j = 2*i+1;
						for( h=0; h < m; h++) {
							if(j < 0) j += n;
							_VT_Range _j(j);
							result(_i ) += hpf(h)*x(_j);
							--j;
						}



					}



				}

				template<class T1, class T2>     
					T1 FWT_PO(  T1 input, int L, T2 qmf) 
				{
					T1 result = _VT_copy(input);
					T2 hpf = MirrorFilter(qmf), lpf = _VT_copy(qmf);

					T1 temp(_VT_shape(result));


					temp =  0;                        
					int  J = (int) ( _VT_log( double( _VT_rows(input))) / _VT_log(2.0) + .5);
					int nj = _VT_rows(input);

					_VT_DeclareRangeAll( _VT_RangeAll);     

					_VT_PRECHECK( _VT_pow(2,J) == nj , 
						"The current implementation of fast wavelet tranform "
						"only supports dyalic lenghs."); 

					for( int j = (J-1); j >= L; --j) {
						_VT_Range q(0,nj-1);

						temp( q) = result(q);

						DownLo(temp,  lpf,  nj,result);
						DownHi(temp,  hpf,  nj,result( _VT_Range(nj/2, nj-1)) );



						nj /=2;
					}

					return result;


				}


				template<class T1, class T2>     
					T1 FWT2_PO(  T1 input, int L, T2 qmf)
				{
					T1 result = _VT_copy(input);
					T2 hpf = MirrorFilter(qmf), lpf = _VT_copy(qmf);

					T1 temp(_VT_shape(result));

					temp =  0;                        
					int  J = (int) ( _VT_log( (double) _VT_rows(input)) / _VT_log(2.0) + .5);
					int nj = _VT_rows(input);

					_VT_PRECHECK( _VT_pow(2,J) == nj , 
						"The current implementation of fast wavelet tranform "
						"only supports dyalic lenghs."); 

					_VT_PRECHECK( _VT_rows(input) == _VT_cols(input),
						"This operation is only valid on square matrices.\n"
						"Invalid shape:" << _VT_shape(input) );

					_VT_DeclareRangeAll(_VT_RangeAll);         

					for( int j = (J-1); j >= L; --j) {
						temp.resize(nj,nj);       
						_VT_Range r_nj(0,nj-1);  
						_VT_Range l_half(nj/2, nj-1);                       
						temp(r_nj,r_nj) = result(r_nj,r_nj);
						DownLo(temp,  lpf,  nj, result(_VT_RangeAll,r_nj));
						DownHi(temp,  hpf,  nj, result(l_half,r_nj));
						result.transposeSelf(1,0);
						temp(r_nj,r_nj) = result(r_nj,r_nj);
						DownLo(temp,  lpf,  nj, result(_VT_RangeAll,r_nj));
						DownHi(temp,  hpf,  nj, result(l_half,r_nj));

						nj /=2;
					}

					result.transposeSelf(1,0);

					return result;


				}



				template<class T1, class T2>
					void UpHi( T1  x, T2 hpf, int n, T1  result)
				{


					int  meven, modd, i, h, j, mmin; 

					/*hipass version */
					int m = _VT_rows(hpf);
					meven = (m+1)/2; 
					modd = m/2;

					/* away from edges */
					_VT_DeclareRangeAll( _VT_RangeAll);

					if(  0 <=  n-meven-1) {
						_VT_Range i_read(0 , n - meven - 1 );
						_VT_Range i_write_1( 1 , 2*(n- meven - 1) + 1, 2);
						_VT_Range i_write_0(  0, 2*(n- meven - 1), 2);


						result( i_write_0 ) = 0;
						result( i_write_1 ) = 0;


						for( h=0; h < meven; h++)
							result(i_write_1) += hpf(2*h)*x(i_read+h);
						for( h=0; h < modd; h++)
							result(i_write_0) += hpf(2*h+1)*x(i_read+h);
					}		

					/* fix up edge values */
					mmin = n-meven;
					if(mmin < 0) mmin = 0;
					for( i= mmin; i<n; i++){
						_VT_Range _ii(2*i);

						result(_ii   ) = 0.;
						result(_ii +1) = 0.;
						j = i;
						for( h=0; h < meven; h++){
							if(j >= n) j -= n;
							_VT_Range _j(j);
							result(_ii + 1) += hpf(2*h)*x(_j);
							j++;
						}

						result(_ii) = 0.;
						j = i;
						for( h=0; h < modd; h++){
							if(j >= n) j -= n;
							_VT_Range _j(j);
							result(_ii ) += hpf(2*h+1)*x(_j);
							j++;
						}



					}

				}

				template<class T1, class T2>
					void UpLo( T1  x, T2 lpf, int n, T1  result)
				{

					int  meven, modd, i, h, j, mmax; 

					int m = _VT_rows(lpf);



					/*lowpass version */

					/* away from edges */
					meven = (m+1)/2; modd = m/2;
					_VT_DeclareRangeAll( _VT_RangeAll );

					if( meven <= n-1) {
						_VT_Range i_read(meven  , n - 1 );


						_VT_Range i_write_1( 1 + 2*meven,  2*(n - 1) + 1, 2);
						_VT_Range i_write_0( 0 + 2*meven,  2*(n - 1), 2);

						result( i_write_0 ) = 0;
						result( i_write_1 ) = 0;

						for( h=0; h < meven; h++)
							result(i_write_0) += lpf(2*h)*x(i_read-h);

						for( h=0; h < modd; h++)
							result(i_write_1) += lpf(2*h+1)*x(i_read-h);

					}


					/* fix up edge values */
					mmax = meven;
					if(mmax > n) mmax = n;
					for( i= 0; i < mmax; i++){
						_VT_Range _ii(2*i);

						result(_ii ) = 0.;
						j = i;
						for( h=0; h < meven; h++){
							if(j < 0) j += n;
							_VT_Range _j(j);
							result(_ii )+= lpf(2*h)*x(_j);
							--j;
						}


						result(_ii  + 1) = 0.;
						j = i;
						for( h=0; h < modd; h++){
							if(j < 0) j += n;
							_VT_Range _j(j);
							result(_ii+ 1) += lpf(2*h+1)*x(_j );
							--j;
						}

					}
				}


				template<class T1, class T2>     
					T1 IWT_PO(  T1 input, int L, T2 qmf) 
				{
					T1 result = _VT_copy(input)  ;
					T2 hpf = MirrorFilter(qmf), lpf = _VT_copy(qmf);
					T1 temp1 = _VT_copy(input), temp2 = _VT_copy(input) ;



					int  J = (int)( _VT_log( (double) _VT_rows(input) ) / _VT_log(2.0) + .5);
					_VT_PRECHECK( _VT_pow(2,J) == _VT_rows(input) , 
						"The current implementation of fast wavelet tranform "
						"only supports dyalic lenghs."); 


					int nl =1; for(int k=0; k < L; k++)  nl *= 2;

					int nj = nl; 



					_VT_DeclareRangeAll( _VT_RangeAll );         

					for(int j=L; j < J; j++) {


						UpLo(result,  lpf,  nj,   temp1);
						UpHi(input( _VT_Range(nj, 2*nj-1) ),  hpf,   nj, temp2);
						nj *=2;
						_VT_Range q(0,nj-1);
						result(q) = temp1(q)+ temp2(q);
					}

					return result;


				}


				template<class T1, class T2>     
					T1 IWT2_PO(  T1 input, int L, T2 qmf) 
				{
					T1 result = _VT_copy(input)  ;
					T2 hpf = MirrorFilter(qmf), lpf = _VT_copy(qmf);
					T1 temp1 = _VT_copy(input), temp2 = _VT_copy(input) ;



					int  J = (int) ( _VT_log( (double) _VT_rows(input) ) / _VT_log(2.0) + .5);

					_VT_PRECHECK( _VT_pow(2,J) == _VT_rows(input) , 
						"The current implementation of fast wavelet tranform "
						"only supports dyalic lenghs."); 

					_VT_PRECHECK( _VT_rows(input) == _VT_cols(input),
						"This operation is only valid on square matrices.\n"
						"Invalid shape:" << _VT_shape(input) );

					int nl =1; for(int k=0; k < L; k++)  nl *= 2;

					int nj = nl; 

					_VT_DeclareRangeAll( _VT_RangeAll );         
					result.transposeSelf(1,0);
					for(int j=L; j < J; j++) {
						// temp.resize(nj,nj);       
						_VT_Range  r_nj(0,2*nj-1);  
						_VT_Range  l_nj(nj, 2*nj-1);                                

						UpLo(result( _VT_RangeAll,r_nj),  lpf,  nj,   temp1( _VT_RangeAll,r_nj) );
						UpHi(result( l_nj,r_nj),  hpf, nj,   temp2( _VT_RangeAll,r_nj) );
						result(r_nj,r_nj) = temp1(r_nj,r_nj)+ temp2(r_nj,r_nj);
						result.transposeSelf(1,0);
						UpLo(result( _VT_RangeAll,r_nj) ,  lpf,  nj,   temp1(_VT_RangeAll,r_nj));
						UpHi(result( l_nj,r_nj), hpf,   nj, temp2( _VT_RangeAll,r_nj));
						result(r_nj,r_nj) = temp1(r_nj,r_nj)+ temp2(r_nj,r_nj);
						nj *=2;



					}

					return result;


				}



			}

		}

	}
}
#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {

	template<class T> 
		T mirrorfilter(T input)
	{
		return _VT_Pack::Functions::Spectral::Wavelet::MirrorFilter(input);
	}
	template<class T1,class T2>	
		T1 downlo(T1 input, T2 lpf)
	{
		//_VT_PRECHECK( _VT_rows(input) must 

		T1 result( _VT_rows(input)/2, _VT_cols(input));
		_VT_Pack::Functions::Spectral::Wavelet::DownLo(input, lpf, 
			_VT_rows(input), result);
		return result;
	}

	template<class T1,class T2>	
		T1 downhi(T1 input, T2 hpf)
	{
		T1 result( _VT_rows(input) /2, _VT_cols(input));
		_VT_Pack::Functions::Spectral::Wavelet::DownHi(input, hpf,
			_VT_rows(input), result);
		return result;
	}

	template<class T1,class T2>	
		T1 uplo(T1 input, T2 lpf)
	{
		T1 result(2*_VT_rows(input), _VT_cols(input));
		_VT_Pack::Functions::Spectral::Wavelet::UpLo(input, lpf, 
			_VT_rows(input), result);
		return result;
	}

	template<class T1,class T2>	
		T1 uphi(T1 input, T2 hpf)
	{
		T1 result(2*_VT_rows(input), _VT_cols(input));
		_VT_Pack::Functions::Spectral::Wavelet::UpHi(input, hpf, 
			_VT_rows(input), 0, result);
		return result;
	}

	template<class T1,class T2> 
		T1 fwt_po(T1 input, int L, T2 qmf) 
	{
		return _VT_Pack::Functions::Spectral::Wavelet::FWT_PO(input, L, qmf);
	} 

	template<class T1,class T2>	
		T1 iwt_po(T1 input, int L, T2 qmf) 
	{
		return _VT_Pack::Functions::Spectral::Wavelet::IWT_PO(input, L, qmf);
	} 

	template<class T1,class T2>	
		T1 iwt2_po(T1 input, int L, T2 qmf) 
	{
		return _VT_Pack::Functions::Spectral::Wavelet::IWT2_PO(input, L, qmf);
	} 

	template<class T1,class T2>	
		T1 fwt2_po(T1 input, int L, T2 qmf) 
	{
		return _VT_Pack::Functions::Spectral::Wavelet::FWT2_PO(input, L, qmf);
	} 


}

#endif // _VT_MATLAB_INTERFACE

#endif // _VT_FUNCTION_FWT_CC
