
/***************************************************************************
* blitz/alg/walsh.cc     
* function implementation of vectorized walsh transform  and
* inverse transform
*
* $Id: walsh.cc,v 1.1.1.1 2003/07/09 23:55:54 mdsuresh Exp $
*
* Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
* 
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* Suggestions:          blitz-dev@oonumerics.org
* Bugs:                 blitz-bugs@oonumerics.org
*
* For more information, please see the Blitz++ Home Page:
*    http://oonumerics.org/blitz/
*
***************************************************************************
* $Log: walsh.cc,v $
* Revision 1.1.1.1  2003/07/09 23:55:54  mdsuresh
*
*
* Revision 1.0  2002/10/26 20:22:49  mdsuresh
* Updated copyright date in headers.
**/

#ifndef _VT_FUNCTION_WALSH_CC
#define _VT_FUNCTION_WALSH_CC

#ifndef _VT_PACK_H
#error the implementation of Walsh transform requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif


namespace _VT_Pack {
	namespace Functions {
		namespace Spectral {
			namespace Walsh {

				template<class T>  
					T Transform( T& input, int s, int n ) 
				{


					_VT_Range a(0,s-1,2);
					_VT_Range b(1,s-1,2);
					_VT_Range first_half(0,(s/2) -1);
					_VT_Range second_half(s/2,s-1);
					_VT_DeclareRangeAll( _VT_RangeAll);

					T result_1( _VT_shape(input) ), result_0 = _VT_copy(input);
					result_1 =  0;

					while(n--) {

						result_1(first_half) = result_0(a) + result_0(b);
						result_1(second_half) = result_0(a) - result_0(b);
						_VT_cycleArrays(result_0, result_1);                  
					}

					return result_0;
				}
				template<class T>
					T Transform( T& input) 
				{
					int s = _VT_rows(input);
					int n = (int) (_VT_log(double(s))  /_VT_log(2.0) + .5);
					//float floor_n = floor(n + .5);
					// assert( n == floor_n);
					_VT_PRECHECK( _VT_pow(2,n) == s , 
						"The current implementation of fast walsh tranform "
						"only supports dyalic lenghs."); 
					return Transform( input, s, (int) n);
				}

				template<class T>
					T InverseTransform( T& input, int s, int n ) 
				{
					if(n  == 0)
						return input;

					_VT_PRECONDITION( n >= 0);
					_VT_PRECONDITION( s >= 0);


					_VT_Range a(0,s-1,2);
					_VT_Range b(1,s-1,2);
					_VT_Range first_half(0,s/2-1);
					_VT_Range second_half(s/2,s-1);
					_VT_DeclareRangeAll( _VT_RangeAll );

					T result_1( _VT_shape(input)), result_0 = _VT_copy(input);
					result_1 = 0;

					while(n-- > 0) {                                      
						result_1(a) = 0.5*(result_0(first_half) + result_0(second_half));
						result_1(b) = 0.5*(result_0(first_half) - result_0(second_half));
						_VT_cycleArrays(result_0, result_1);                   

					}

					return  result_0;
				}

				template<class T>
					T InverseTransform( T& input) 
				{
					int s = _VT_rows(input);
					int n = int(_VT_log( double(s))  /_VT_log(2.0) + .5);
					_VT_PRECHECK( _VT_pow(2,n) == s , 
						"The current implementation of fast walsh tranform "
						"only supports dyalic lenghs."); 
					//float floor_n = floor(n + .5);
					//assert( n == floor_n);


					return InverseTransform( input, s, n);
				}



			}

		}
	}

}

#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {
	template<class T>
		T walsh( T & input)
	{
		return _VT_Pack::Functions::Spectral::Walsh::Transform(input);
	}
	template<class T>
		T iwalsh(T & input)
	{ 
		return _VT_Pack::Functions::Spectral::Walsh::InverseTransform(input);
	}
}

#endif // _VT_MATLAB_INTERFACE

#endif // _VT_FUNCTION_WALSH_CC
