/***************************************************************************
 * blitz/xlistint.h    Definition of the extended classes for 
 * initialization lists
 *
 * $Id: xlistinit.cc,v 1.1.1.1 2003/07/09 23:55:51 mdsuresh Exp $
 *
 * Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Suggestions:          blitz-dev@oonumerics.org
 * Bugs:                 blitz-bugs@oonumerics.org
 *
 * For more information, please see the Blitz++ Home Page:
 *    http://oonumerics.org/blitz/
 *
 ***************************************************************************
 * $Log: xlistinit.cc,v $
 * Revision 1.1.1.1  2003/07/09 23:55:51  mdsuresh
 *
 *
 * Revision 1.0
 * 
 * These file exists only as proof of concept. Once this feature is found
 * useful, it would not longer by refered to by the eXtented name. 
 */

#ifndef BZ_xLISTINIT_H
#define BZ_xLISTINIT_H

BZ_NAMESPACE(blitz)

template<char c>
class delim
{
	enum {
		value = c
	};
};


template<class T_array, class T_iterator>
class _xListInitializer;

template<class T_array, class T_iterator>
class _xListWorker;

template<int N_rank_over_2, class T_array, class T_iterator>
struct _xSelectAppropriateListInitializerSwitch {

	   typedef ListInitializationSwitch < T_array,  T_iterator*> T_ListSwitch;
	   enum  {
	      matConcat = 0
	   };
};

template< class T_array, class T_iterator>
struct _xSelectAppropriateListInitializerSwitch< 0 , T_array, T_iterator> {
	   
	typedef _xListWorker< T_array,  T_iterator> T_ListSwitch;
    enum {
		  matConcat = 1
	};
};

class _xUnknownUnusableType {

};

template<bool _true, class T_numtype, int rank>
struct _xDefineProperArrayParam
{
   typedef Array< T_numtype , rank> T_properarray;
};

template<class T_numtype, int rank>
struct _xDefineProperArrayParam< 0, T_numtype, rank>
{
   typedef Array< _xUnknownUnusableType , rank> T_properarray;
};

    template<class T>
	void SetArrayValueAndMove( Array<T, 1>&  a, int& col, T val)
	{ 
	   a(col++) = val;
	}
	
	template<class T>
	void SetArrayValueAndMove( Array<T,2> & a, int& col, T val)
	{ 
	   a(0,col++) = val;
	}
	template<class T>
	T getArrayValue( Array<T, 1>  a, int col)
	{ 
	   return a(col) ;
	}
	
	template<class T>
	T getArrayValue( Array<T,2>  a, int col)
	{ 
	   return a(0,col);
	}
	
	template<class T>
	int getCols( Array<T, 1>  a)
	{ 
	   return a.rows() ;
	}
	
	template<class T>
	int getCols( Array<T,2>  a)
	{ 
	   return a.cols();
	}
	
	template<class T>
	int getRows( Array<T, 1>  a)
	{ 
	   return 1 ;
	}
	
	template<class T>
	int getRows( Array<T,2>  a)
	{ 
	   return a.rows();
	}




template<class T_array, class T_ierator>
class _xListWorkerEnd {
	   
};



template<class T_array, class T_iterator>
class _xListWorker {
	typedef _bz_typename T_array::T_numtype T_numtype;
	
	enum {
     _Rank_G0 =	  T_array::_bz_rank > 0,
     _Rank_G1 =   T_array::_bz_rank > 1,
     _Rank_G2 =   T_array::_bz_rank > 2
	};
	

public:
	
	_xListWorker(T_array& array, delim<'['> _begin): 
	array_(array)
	{
		col_ = 0;
		arows_ = -1;
    }



	
	~_xListWorker()
	{
	   // Compiler Should Error here, since the array has not been closed.
       // int * q = static_cast< int *>(double(1));

	  //if(wipeondestruct)
	  //       array_.initialize( value_ );
	}
	

	void Write(const T_numtype& x)
	{
		BZPRECHECK( arows_ == 1 || arows_ == -1, 
			"Concation failed because of matrix shape mismatch.\n" 
			"Expected shape: [" << arows_  << ", x ].\n"
			"Got shape: [ 1 , 1 ] \n");
			
		//cout << array_.shape();

		SetArrayValueAndMove(array_, col_ ,x);
		
		if( (col_ == getCols( array_ )) && (getRows(array_) > 1)  ) 
		  LineFeed(1);
		else
			arows_ = 1;
	}

	void Write( const _bz_typename _xDefineProperArrayParam< 
							    _Rank_G0 ,  
							    T_numtype,1>::T_properarray & x)
	{
		if( _Rank_G0 ) {
			BZPRECHECK( arows_ == -1 ||  arows_ == 1
				, "Concat failed because of matrix shape mismatch.\n" 
				"Expected shape: [" << arows_  << ", x ].\n"
				"Got shape: [ 1 ," << x.rows() << "] \n");

			array_( 0 , -1 + col_ + Range(1,x.rows(),1)  ) = x;
			if( ((col_ += x.rows()) == getCols( array_ )) && (getRows(array_) > 1) ) 
			  LineFeed(1);
			else
				arows_ = 1;
		}
	}

	void Write(const  _bz_typename _xDefineProperArrayParam< 
				_Rank_G1 , 
				T_numtype,2>::T_properarray & x)
	{
		if (_Rank_G1 ) {

			BZPRECHECK( arows_ == -1 || arows_ == x.rows()
				, "Concat failed because of matrix shape mismatch.\n" 
				"Expected shape: [" << arows_  << ", x ].\n"
				"Got shape: [ 1 ," << x.rows() << "] \n");

			array_( -1  + 0 + Range(1,x.rows(),1) ,
				-1  + col_ + Range(1,x.cols(),1) ) = x;


			if( ((col_ += x.cols()) == getCols( array_)) &&  (getRows(array_) > 1) ) 
			   LineFeed(arows_);
			else
				arows_ = x.rows();

		}

	}


	void LineFeed( int rows )
	{
			array_.slice(0, Range( rows, Range::toEnd) );
			col_=0;
			arows_ = -1;
	}

	template<class _T>
		    _xListWorker<T_array, T_iterator>& operator,(_T x)
	{
	    Write(x);
		return *this;
	}

    _xListWorker<T_array, T_iterator>& operator,( delim<';'>  x)
	{
		array_( Range(0, arows_-1) , Range( col_, Range::toEnd) ) = 0;
	 	LineFeed(arows_);
		return *this;
	}

    _xListWorkerEnd<T_array,T_iterator> operator,( delim<']'>  x)
	{
		return _xListWorkerEnd<T_array,T_iterator>();
	}
	

private:
	_xListWorker()
	{

	}

protected:
	T_array array_;
	int arows_;
	int col_;
};


BZ_NAMESPACE_END

#endif // BZ_LISTINIT_H
